
/*
 * Created on 26 avr. 2004
 *
 */

package fr.umlv.sprintabout;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;

import static org.objectweb.asm.Opcodes.*;

/**
 * @author remi
 *
 */
public class VisitorGenerator {
  
  public <T> T createVisitor(Class<T> visitorClass) {
    try {
      return getClass(visitorClass).newInstance();
    } catch(InstantiationException e) {
      throw new AssertionError(e);
    } catch(IllegalAccessException e) {
      throw new AssertionError(e);
    }
  }
  
  @SuppressWarnings("unchecked")
  private <T> Class<T> getClass(Class<T> visitorClass) {
    String visitorClassName=visitorClass.getName();
    String stubClassName=visitorClassName+"$_$Visitor";
    
    // use system class loader as a cache
    try {
      return (Class<T>)ClassLoader.getSystemClassLoader().loadClass(stubClassName);
    } catch(ClassNotFoundException e) {
      // do nothing
    }
    
    byte[] code=generate(visitorClass,stubClassName.replace('.','/'));
    
    /* DEBUG 
    try {
      String debugClassName=stubClassName.substring(stubClassName.lastIndexOf('.')+1);
      java.io.FileOutputStream output=new java.io.FileOutputStream(debugClassName+".class");
      output.write(code);
      output.flush();
    } catch(java.io.IOException e) {
      e.printStackTrace();
    }
    */
    
    return (Class<T>)createClass(stubClassName,code);
  }
  
  private static Class<?> createClass(String className,byte[] code) {
    try {
      return (Class<?>) defineClass.invoke(ClassLoader.getSystemClassLoader(),
        className,code,0,code.length);
    } catch(IllegalAccessException e) {
      throw new AssertionError(e);
    } catch(InvocationTargetException e) {
      throw new AssertionError(e.getCause());
    }
  }
  
  private static final Method defineClass;
  
  static {
    try {
      defineClass=ClassLoader.class.getDeclaredMethod("defineClass",
        String.class, byte[].class, int.class, int.class );
    } catch(NoSuchMethodException e) {
      throw new AssertionError(e);
    }
    defineClass.setAccessible(true);
  }
  
  private byte[] generate(Class<?> visitorClass,String internalClassName) {
    
    String visitorClassName=visitorClass.getName();
      
    // JDK1.5 class file
    ClassWriter writer=new ClassWriter(0);
    writer.visit(V1_5, ACC_PUBLIC, internalClassName,null,visitorClassName.replace('.','/'), null);
    
    // static init
    MethodVisitor blockVisitor=writer.visitMethod(ACC_STATIC,"<clinit>", "()V", null, null);
    
    // public constructor
    generateConstructor(writer,visitorClassName);
    
    // generate selector
    Map<Selector,List<Method>> selectors=gatherSelectors(visitorClass);
    for(Map.Entry<Selector,List<Method>> entry:selectors.entrySet()) {
      
      Selector selector=entry.getKey();
      Algorithm algo=factory.getAlgorithm(selector,entry.getValue());
      
      generateSelector(writer,internalClassName,blockVisitor,selector,algo);
    }
   
    writer.visitEnd();
    
    return writer.toByteArray();
  }
  
  private void generateConstructor(ClassWriter writer,String visitorClassName) {
    MethodVisitor visitor=writer.visitMethod(ACC_PUBLIC,"<init>","()V", null, null);
    visitor.visitVarInsn(ALOAD,0);
    visitor.visitMethodInsn(INVOKESPECIAL,visitorClassName.replace('.','/'),
      "<init>","()V");
    visitor.visitInsn(RETURN);
    visitor.visitMaxs(1,1);
  }
  
  private void generateSelector(ClassWriter writer,String internalClassName,MethodVisitor blockVisitor,Selector selector,Algorithm algo) {
    // generate current hash maps
    for(int i=0;i<selector.length;i++) {
      writer.visitField(ACC_STATIC|ACC_PRIVATE|ACC_FINAL,selector.getMapFieldName(i),
        "Lfr/umlv/sprintabout/ConcurrentIntMap;",null,null);
    }
    
    generateCall(writer,internalClassName,selector);
    
    generateStaticBlock(writer,internalClassName,blockVisitor,selector,algo);
    
    generateDispatch(writer,internalClassName,selector,algo);
  }
  
  private void generateStaticBlock(ClassWriter writer,String internalClassName,MethodVisitor visitor,Selector selector,Algorithm algo) {
    
    int argumentCount=algo.getArgumentCount();
    for(int argumentIndex=0;argumentIndex<argumentCount;argumentIndex++) {
      visitor.visitTypeInsn(NEW,"fr/umlv/sprintabout/ConcurrentIntMap");
      visitor.visitInsn(DUP);
      
      pushInt(visitor,algo.getMethodCount()<<1);
      visitor.visitMethodInsn(INVOKESPECIAL,"fr/umlv/sprintabout/ConcurrentIntMap",
        "<init>","(I)V");
      
      //TODO need to push null value
      
      int methodCount=algo.getMethodCount();
      for(int methodIndex=0;methodIndex<methodCount;methodIndex++) {
        visitor.visitInsn(DUP);
        
        Class<?> type=algo.getTypeAt(methodIndex,argumentIndex);
        int annotation=algo.getAnnotationAt(methodIndex,argumentIndex);
        
        // LDC with Class
        visitor.visitLdcInsn(Type.getType(type));
        
        //visitor.visitLdcInsn(annotation);
        pushInt(visitor,annotation);
        
        visitor.visitMethodInsn(INVOKEVIRTUAL,"fr/umlv/sprintabout/ConcurrentIntMap",
          "insert","(Ljava/lang/Class;I)V");
      }
      
      visitor.visitFieldInsn(PUTSTATIC,internalClassName,
        selector.getMapFieldName(argumentIndex), "Lfr/umlv/sprintabout/ConcurrentIntMap;");
    }
    visitor.visitInsn(RETURN);
    visitor.visitMaxs(4,0);
  }
    
  private void generateCall(ClassWriter writer,String internalClassName,Selector selector) {
    Method selectorMethod=selector.method;
    Type[] types=Type.getArgumentTypes(selectorMethod);
    
    String methodDescriptor=Type.getMethodDescriptor(Type.getReturnType(selectorMethod),types);
    
    MethodVisitor visitor=writer.visitMethod(ACC_PUBLIC,selectorMethod.getName(),
      Type.getMethodDescriptor(selectorMethod), null, null);
    
    visitor.visitVarInsn(ALOAD,0);
    
    for(int i=0;i<types.length;i++) {
      visitor.visitFieldInsn(GETSTATIC,internalClassName,
        selector.getMapFieldName(i),"Lfr/umlv/sprintabout/ConcurrentIntMap;");
      visitor.visitVarInsn(types[i].getOpcode(ILOAD),i+1);
      visitor.visitMethodInsn(INVOKESTATIC,"fr/umlv/sprintabout/VisitorGenerator",
        "getFromMap","(Lfr/umlv/sprintabout/ConcurrentIntMap;Ljava/lang/Object;)I");
      if (i!=0)
        visitor.visitInsn(IADD);
    }
    
    for(int i=0;i<types.length;i++) {
      visitor.visitVarInsn(types[i].getOpcode(ILOAD),i+1);
    }
    visitor.visitMethodInsn(INVOKEVIRTUAL,internalClassName,
      selector.getDispatchName(),"(I"+methodDescriptor.substring(1));
    
    visitor.visitInsn(Type.getReturnType(selectorMethod).getOpcode(IRETURN));
    
    //System.out.println(Math.max((types.length==1)?3:4,2+types.length));
    visitor.visitMaxs(Math.max((types.length==1)?3:4,2+types.length),1+types.length);
  }
  
  private void generateDispatch(ClassWriter writer,String internalClassName,Selector selector,Algorithm algo) {
    Method selectorMethod=selector.method;
    Type[] types=Type.getArgumentTypes(selectorMethod);
    
    String methodDescriptor=Type.getMethodDescriptor(Type.getReturnType(selectorMethod),types);
    
    MethodVisitor visitor=writer.visitMethod(ACC_PRIVATE,selector.getDispatchName(),
      "(I"+methodDescriptor.substring(1), null, null);
    
    Label error=new Label();

    int methodCount=algo.getMethodCount();
    
    Label[] labels=new Label[methodCount];
    for(int i=0;i<methodCount;i++)
      labels[i]=new Label();
    
    visitor.visitVarInsn(ILOAD,1);
    
    
    visitor.visitTableSwitchInsn(1,methodCount,error,labels);
    
    
    /*
    for(int i=0;i<labels.length;i++) {
      if (i!=labels.length-1)
        visitor.visitInsn(DUP);
      
      visitor.visitLdcInsn(i+1);
      visitor.visitJumpInsn(IF_ICMPEQ,labels[i]);
    }
    //visitor.visitJumpInsn(GOTO,error);
    */
    
    visitor.visitLabel(error);
    visitor.visitTypeInsn(NEW,"java/lang/AssertionError");
    visitor.visitInsn(DUP);
    visitor.visitLdcInsn("bad index, ooooooooooups");
    visitor.visitMethodInsn(INVOKESPECIAL,"java/lang/AssertionError",
      "<init>","(Ljava/lang/Object;)V");
    visitor.visitInsn(ATHROW);
    
    for(int dispatchIndex=0;dispatchIndex<methodCount;dispatchIndex++) {
      visitor.visitLabel(labels[dispatchIndex]);
      
      Method impl=algo.getMethodAtDispatchIndex(dispatchIndex);
      
      boolean isStatic=Modifier.isStatic(impl.getModifiers());
      if (!isStatic)
        visitor.visitVarInsn(ALOAD,0);
      
      Class<?>[] implTypes=impl.getParameterTypes();
      for(int j=0;j<implTypes.length;j++) {
        visitor.visitVarInsn(types[j].getOpcode(ILOAD),j+2);
        visitor.visitTypeInsn(CHECKCAST,Type.getInternalName(implTypes[j]));
      }
 
      visitor.visitMethodInsn(
        (isStatic)?INVOKESTATIC:INVOKEVIRTUAL,
        Type.getInternalName(impl.getDeclaringClass()),
        impl.getName(),
        Type.getMethodDescriptor(impl));
      visitor.visitInsn(Type.getReturnType(impl).getOpcode(IRETURN));
    }
    
    
    visitor.visitMaxs(Math.max(3,types.length),2+types.length);
  }
  
  
  
  private final AlgorithmFactory factory=new AlgorithmFactory();
  
  private static void pushInt(MethodVisitor visitor,int value) {
    if (value>=-1 && value<=5)
      visitor.visitInsn(ICONST_M1+value+1);
    else
      if (value>=Byte.MIN_VALUE && value<=Byte.MAX_VALUE)
        visitor.visitIntInsn(BIPUSH,value);
      else
        if (value>=Short.MIN_VALUE && value<=Short.MAX_VALUE)
          visitor.visitIntInsn(SIPUSH,value);
        else
          visitor.visitLdcInsn(value);
  }
  
  private static Map<Selector,List<Method>> gatherSelectors(Class<?> visitorClass) {
    // two passes algorithm
    Method[] methods=visitorClass.getMethods();
    
    // first find selectors
    HashMap<Selector,List<Method>> selectors=new HashMap<Selector,List<Method>>();
    for(Method method:methods) {
      if (!method.getName().endsWith("Appropriate") || !Modifier.isAbstract(method.getModifiers()))
        continue;
      
      Selector selector=new Selector(method);
      if (selectors.put(selector,new ArrayList<Method>())!=null)
        throw new IllegalStateException("two method visitAppropriate with same number of parameter");
        
    }
    
    // second pass, map method to selector
    for(Method method:methods) {
      List<Method> list=selectors.get(new Selector(method.getName(),method.getParameterTypes().length));
      if (list==null)
        continue;
      
      list.add(method);
    }
    
    return selectors;
  }
  
  
  
  public static int getFromMap(ConcurrentIntMap map,Object o) {
    Class<?> clazz=(o==null)?void.class:o.getClass();
    int ann=map.get(clazz);
    if (ann==-1) {
      ann=processFromMap(map,clazz);
      map.insert(clazz,ann);
    }
    return ann;
  }
  
  public static int processFromMap(ConcurrentIntMap map,Class<?> clazz) {
    Class<?> superclazz=clazz.getSuperclass();
    int ann;
    if (superclazz!=null) { 
      ann=map.get(superclazz);
      if (ann==-1)
        ann=processFromMap(map,superclazz); 
    } else {
      ann=Integer.MIN_VALUE;
    }
    
    Class<?>[] interfazes=clazz.getInterfaces();
    for(int i=0;i<interfazes.length;i++) {
      Class<?> interfaze=interfazes[i];
      int ann2=map.get(interfaze);
      if (ann2==-1)
        ann2=processFromMap(map,interfaze);
      
      ann=Math.max(ann,ann2);
    }
    return ann;
  }
}
